#!/usr/bin/env python3

"""
code taken from a project at Stanford CS106A
"""

import tkinter
import random
import datetime

from byugrid import Grid


def do_move(grid, x_from, y_from, x_to, y_to):
    """
    Move the sand from (x_from, y_from) to (x_to, y_to) and return the
    resulting grid. Assume that this is a legal move: all coordinates are
    in bounds, and (x_to, y_to) is empty.
    (a different function checks that this is a legal move before do_move() is called)
    :param grid: a grid with rocks and sand
    :param x_from: x coordinate of a cell with sand in it
    :param y_from: y coordinate of a cell with sand in it
    :param x_to: x coordinate of a cell to move the sand to
    :param y_to: y coordinate of a cell to move the sand to
    :return a modified grid
    >>> grid1 = Grid.build([['r', 's', 's'], [None, None, None]])
    >>> do_move(grid1, 1, 0, 1, 1)
    [['r', None, 's'], [None, 's', None]]
    >>>
    >>> grid2 = Grid.build([['r', 's', 's'], [None, None, None]])
    >>> do_move(grid2, 2, 0, 2, 1)
    [['r', 's', None], [None, None, 's']]
    """
    # Erase where it was, set it to its new square
    grid.set(x_from, y_from, None)
    grid.set(x_to, y_to, 's')
    return grid


def is_move_ok(grid, x_from, y_from, x_to, y_to):
    """
    Check if it is possible to move a piece of sand from (x_from, y_from)
    to (x_to, y_to). Return True if this is possible, False otherwise.

    Assume there is sand at (x_from, y_from) and this location is in bounds. For
    a move to be OK:

    (1) the destination must be in bounds,
    (2) the destination must be empty, and
    (3) the move must not violate the corner rule.

    The corner rule states that for a down-right or down-left move, the
    cell above the destination must be empty.
    :param grid: a grid with rocks and sand
    :param x_from: x coordinate of a cell with sand in it
    :param y_from: y coordinate of a cell with sand in it
    :param x_to: x coordinate of a cell to move the sand to
    :param y_to: y coordinate of a cell to move the sand to
    :return True or False
    >>> # Out-of-bounds (OOB) tests
    >>> # Make a 1 by 1 grid with an 's' - everything is OOB from here
    >>> grid1 = Grid.build([['s']])
    >>> is_move_ok(grid1, 0, 0, -1, 0) # left OOB
    False
    >>> is_move_ok(grid1, 0, 0, 0, 1)  # down OOB
    False
    >>> is_move_ok(grid1, 0, 0, -1, 1)  # down-left OOB
    False
    >>> # TBD - implement 2 more OOB tests here
    >>> # 3 by 2 grid, try various moves from 1,0
    >>> grid2 = Grid.build([[None, 's',   'r'], [None, None, None]])
    >>> is_move_ok(grid2, 1, 0, 0, 0)  # left ok
    True
    >>> # TBD - implement 4 more tests here
    """
    pass


def do_gravity(grid, x, y):
    """
    Given a grid and a coordinate (x, y) that is in bounds, check if there is
    sand at this coordinate. If there is, try to make one move, trying them
    in this order:

    (1) move down, (2) move down-left, (3) move down-right.

    Return the grid in all cases.
    :param grid: a grid with rocks and sand
    :param x: x coordinate to check
    :param y: y coordinate to check
    :return a modified grid
    >>> # not sand
    >>> grid1 = Grid.build([[None, 's', None], [None, None, None]])
    >>> do_gravity(grid1, 0, 0)
    [[None, 's', None], [None, None, None]]
    >>>
    >>> # down
    >>> grid2 = Grid.build([[None, 's', None], [None, None, None]])
    >>> do_gravity(grid2, 1, 0)
    [[None, None, None], [None, 's', None]]
    >>>
    >>> # bottom blocked
    >>> grid3 = Grid.build([[None, 's', None], ['r', 'r', 'r']])
    >>> do_gravity(grid3, 1, 0)
    [[None, 's', None], ['r', 'r', 'r']]
    >>>
    >>> # rock-below down-left
    >>> grid4 = Grid.build([[None, 's', None], [None, 'r', None]])
    >>> do_gravity(grid4, 1, 0)
    [[None, None, None], ['s', 'r', None]]
    >>>
    >>> # sand-below down-right
    >>> grid5 = Grid.build([[None, 's', None], ['s', 's', None]])
    >>> do_gravity(grid5, 1, 0)
    [[None, None, None], ['s', 's', 's']]
    >>>
    >>> # sand corner: down-right
    >>> grid6 = Grid.build([['s', 's', None], [None, 's', None]])
    >>> do_gravity(grid6, 1, 0)
    [['s', None, None], [None, 's', 's']]
    >>>
    >>> # at bottom already
    >>> grid7 = Grid.build([[None, None, None], [None, 's', None]])
    >>> do_gravity(grid7, 1, 1)
    [[None, None, None], [None, 's', None]]
    """
    pass


def do_brownian(grid, x, y, brownian):
    """
    Given a grid, a coordinate (x, y), and a brownian integer,
    do the random brownian move for that coordinate (x, y)

    :param grid: a grid with rocks and sand
    :param x: x coordinate to check
    :param y: a y coordinate to check
    :param brownian: a brownian integer from 0 to 100
    :return a modified grid
    >>> # Hack: tamper with randrange() to always return 0
    >>> # So we can write a test.
    >>> # This only happens for the Doctest run, not in any other case.
    >>> random.randrange = lambda n: 0
    >>> # 1,0 can go left, but 1,1 cannot
    >>> grid1 = Grid.build([[None, 's', None], ['s', 's', None]])
    >>> do_brownian(grid1, 1, 0, 100)
    [['s', None, None], ['s', 's', None]]
    >>> grid2 = Grid.build([[None, 's', None], ['s', 's', None]])
    >>> do_brownian(grid2, 1, 1, 100)
    [[None, 's', None], ['s', 's', None]]
    """
    # check if there is sand here
    pass


def do_whole_grid(grid, brownian):
    """
    Do one round of gravity and brownian over the whole grid.

    :param grid: a grid of rock and sand
    :brownian: a brownian integer
    """
    pass


#########################################################

"""
Down here is not especially pretty code to set up the GUI,
handle the controls, and draw the grid to the screen.

Don't write any code below here.

"""


def draw_grid_canvas(grid, canvas, scale):
    """
    Draw grid to tk canvas, erasing and then filling it.
    This was ultimately the best performing approach.
    scale is pixels per block
    """
    # pixel size of canvas
    c_width = grid.width * scale + 2
    c_height = grid.height * scale + 2

    canvas.delete('all')

    # draw black per spot
    for y in range(grid.height):
        for x in range(grid.width):
            val = grid.get(x, y)
            if val:
                if val == 'r':
                    color = 'black'
                else:
                    color = 'yellow'
                rx = 1 + x * scale
                ry = 1 + y * scale
                canvas.create_rectangle(rx, ry, rx + scale, ry + scale, fill=color, outline='black')

    canvas.create_rectangle(0, 0, c_width-1, c_height-1, outline='blue')
    canvas.update()


fps_enable = True
fps_count = 0
fps_start = 0


def fps_update():
    global fps_enable, fps_count, fps_start, fps_label
    if not fps_enable:
        return
    fps_count += 1
    if fps_count == 40:
        now = datetime.datetime.now().timestamp()
        delta = now - fps_start
        fps_start = now
        fps = int(1 / (delta / fps_count))
        # print(fps)
        fps_label.config(text=str(fps))
        fps_count = 0


# Global pointers to GUI elements we need in various callbacks
gravity = None
content = None
brownian_on = None
brownian_val = None
fps_label = None

SHIFT = 6


# provided function to build the GUI
def make_gui(top, width, height):
    """
    Set up the GUI elements for the Sand window, returning the Canvas to use.
    top is TK root, width/height is canvas size.
    """

    global gravity, content, brownian_on, brownian_val, fps_label
    gravity = tkinter.IntVar()
    content = tkinter.StringVar()
    brownian_on = tkinter.IntVar()
    brownian_val = tkinter.IntVar()

    top.title('Sand')

    # gravity checkbox
    gcheck = tkinter.Checkbutton(top, text='Gravity', name='gravity', variable=gravity)
    gcheck.grid(row=0, column=0, sticky='w')
    gravity.set(1)

    scheck = tkinter.Checkbutton(top, text='Brownian', name='brownian', variable=brownian_on)
    scheck.grid(row=0, column=1, sticky='w')
    brownian_on.set(1)

    scale = tkinter.Scale(top, from_=0, to=100, orient=tkinter.HORIZONTAL, variable=brownian_val)
    scale.grid(row=0, column=2, sticky='w')
    brownian_val.set(20)

    # content variable = state of radio button
    sand = tkinter.Radiobutton(top, text="Sand", variable=content, value='s')
    sand.grid(row=0, column=3, sticky='w')

    rock = tkinter.Radiobutton(top, text="Rock", variable=content, value='r')
    rock.grid(row=0, column=4, sticky='w')

    erase = tkinter.Radiobutton(top, text="Erase", variable=content, value='erase')
    erase.grid(row=0, column=5, sticky='w')

    bigerase = tkinter.Radiobutton(top, text="BigErase", variable=content, value='bigerase')
    bigerase.grid(row=0, column=6, sticky='w')

    content.set('s')

    fps_label = tkinter.Label(top, text="0", fg='lightgray')  # ugh 'fg' not a great name for this!
    fps_label.grid(row=0, column=7, sticky='w')

    # canvas for drawing
    canvas = tkinter.Canvas(top, width=width, height=height, name='canvas')

    canvas.xview_scroll(SHIFT, "units")  # hack so (0, 0) works correctly
    canvas.yview_scroll(SHIFT, "units")

    canvas.grid(row=1, columnspan=12, sticky='w', padx=20, ipady=5)

    top.update()
    return canvas


def big_erase(grid, x, y, canvas, scale):
    """Erase big red circle in the given grid centered on x,y"""
    rad = 4
    # Compute circle around x,y in grid coords
    x1 = x - rad  # this can be out of bounds
    y1 = y - rad

    x2 = x + rad
    y2 = y + rad

    # Draw a red circle .. will be erased by later updates
    # Need to be consistent about grid -> pixel mapping
    canvas.create_oval(1 + x1 * scale, 1 + y1 * scale, 1 + x2 * scale, 1 + y2 * scale,
                       fill='red', outline='')
    canvas.update()

    for ey in range(y1, y2 + 1):
        for ex in range(x1, x2 + 1):
            # circle around x,y
            if grid.in_bounds(ex, ey) and abs(x-ex)**2 + abs(y-ey)**2 <= rad ** 2:
                grid.set(ex, ey, None)


def start_timer(top, speed, fn):
    """Start the my_timer system, calls given fn"""
    top.after(speed, lambda: my_timer(top, speed, fn))


def my_timer(top, speed, fn):
    """my_timer callback, re-posts itself."""
    fn()
    top.after(speed, lambda: my_timer(top, speed, fn))


def sand_action(grid, canvas, scale):
    """This function runs on timer for all periodic tasks."""
    global gravity
    global mouse_fn
    global brownian_on
    global brownian_val

    if mouse_fn:
        mouse_fn()

    if gravity.get():
        if not brownian_on.get():
            val = 0
        else:
            val = brownian_val.get()
        do_whole_grid(grid, val)
    draw_grid_canvas(grid, canvas, scale)
    fps_update()


# global mouse sand_action function pointer
# set on mouse down, cleared on mouse-up
mouse_fn = None


def do_mouse_up(event):
    global mouse_fn
    mouse_fn = None


def do_mouse(event, grid, scale, canvas):
    """Callback for mouse click/move"""
    global mouse_fn
    mouse_fn = lambda: do_mouse(event, grid, scale, canvas)

    x = (event.x - SHIFT // 2) // scale
    y = (event.y - SHIFT // 2) // scale
    if grid.in_bounds(x, y):
        global content
        val = content.get()  # 's' 'r' None
        if val == 's' or val == 'r':
            grid.set(x, y, val)
        elif val == 'erase':
            grid.set(x, y, None)
        elif val == 'bigerase':
            big_erase(grid, x, y, canvas, scale)
    # print('click', event.x, event.y)


# (provided)
def main():
    import argparse

    # setup the argument parser
    parser = argparse.ArgumentParser(description="Run Conway's game of life")
    parser.add_argument('--width', type=int, default=50, help='width of the board, 50 by default')
    parser.add_argument('--height', type=int, default=50, help='height of the board, 50 by default')
    parser.add_argument('--speed', type=int, default=1, help='speed of each round, 1ms by default')
    parser.add_argument('--scale', type=int, default=10, help='scale of board, 10 by default')

    # parse the command line arguments
    args = parser.parse_args()

    top = tkinter.Tk()
    canvas = make_gui(top, args.width * args.scale + 2, args.height * args.scale + 2)
    grid = Grid(args.width, args.height)

    canvas.bind("<B1-Motion>", lambda evt: do_mouse(evt, grid, args.scale, canvas))
    canvas.bind("<Button-1>", lambda evt: do_mouse(evt, grid, args.scale, canvas))
    canvas.bind("<ButtonRelease-1>", lambda evt: do_mouse_up(evt))

    start_timer(top, args.speed, lambda: sand_action(grid, canvas, args.scale))

    tkinter.mainloop()


if __name__ == '__main__':
    main()
